from queue import Empty

import pytest

from smyth.exceptions import LambdaHandlerLoadError, LambdaTimeoutError
from smyth.runner.fake_context import FakeLambdaContext
from smyth.runner.process import RunnerProcess
from smyth.types import SmythHandlerState

pytestmark = pytest.mark.anyio


@pytest.fixture
def mock_setproctitle(mocker):
    return mocker.patch("smyth.runner.process.setproctitle")


@pytest.fixture
def mock_logging_dictconfig(mocker):
    return mocker.patch("logging.config.dictConfig")


@pytest.fixture
def runner_process():
    return RunnerProcess("test_process", "tests.conftest.example_handler")


def test_init_process(runner_process):
    assert runner_process.name == "test_process"
    assert runner_process.task_counter == 0
    assert runner_process.last_used_timestamp == 0
    assert runner_process.state == SmythHandlerState.COLD
    assert runner_process.lambda_handler_path == "tests.conftest.example_handler"
    assert runner_process.log_level == "INFO"


def test_stop_process(runner_process):
    runner_process.start()
    assert runner_process.is_alive() is True
    runner_process.stop()
    assert runner_process.is_alive() is False


@pytest.mark.skip(reason="This needs more thought")
def test_send_process(runner_process):
    pass


def test_run(mocker, mock_setproctitle, mock_logging_dictconfig, runner_process):
    mock_lambda_invoker__ = mocker.patch.object(
        runner_process, "lambda_invoker__", autospec=True
    )
    mock_get_logging_config = mocker.patch(
        "smyth.runner.process.get_logging_config", autospec=True
    )
    runner_process.run()
    mock_setproctitle.assert_called_once_with(f"smyth:{runner_process.name}")
    mock_logging_dictconfig.assert_called_once_with(
        mock_get_logging_config.return_value
    )
    mock_lambda_invoker__.assert_called_once()


def test_get_message(mocker, runner_process):
    mock_input_queue = mocker.patch.object(runner_process, "input_queue", autospec=True)
    mock_input_queue.get.side_effect = [
        {"type": "smyth.lambda.invoke", "event": {}, "context": {}},
        Empty,
        {"type": "smyth.lambda.invoke", "event": {}, "context": {}},
        {"type": "smyth.stop"},
    ]

    messages = list(runner_process.get_message__())

    assert len(messages) == 2
    assert messages[0]["type"] == "smyth.lambda.invoke"
    assert messages[1]["type"] == "smyth.lambda.invoke"


def test_get_event(mocker, runner_process):
    assert (
        runner_process.get_event__({"type": "smyth.lambda.invoke", "event": {}}) == {}
    )


def test_get_context(mocker, runner_process):
    assert isinstance(
        runner_process.get_context__({"type": "smyth.lambda.invoke", "context": {}}),
        FakeLambdaContext,
    )

    assert (
        runner_process.get_context__(
            {"type": "smyth.lambda.invoke", "context": {}}
        ).timeout
        == 6
    )


def test_import_handler(mocker, runner_process):
    mock_import_attribute = mocker.patch(
        "smyth.runner.process.import_attribute", autospec=True
    )

    runner_process.import_handler__("tests.conftest.example_handler", {}, {})
    mock_import_attribute.assert_called_once_with("tests.conftest.example_handler")

    mock_import_attribute.side_effect = AttributeError
    with pytest.raises(LambdaHandlerLoadError):
        runner_process.import_handler__("tests.conftest.example_handler", {}, {})

    mock_import_attribute.side_effect = ImportError
    with pytest.raises(LambdaHandlerLoadError):
        runner_process.import_handler__("tests.conftest.example_handler", {}, {})

    mock_import_attribute.side_effect = Exception
    with pytest.raises(Exception):
        runner_process.import_handler__("tests.conftest.example_handler", {}, {})


def test_set_status(runner_process):
    runner_process.set_status__(SmythHandlerState.COLD)
    assert runner_process.state == SmythHandlerState.COLD


def test_timeout_handler(runner_process):
    with pytest.raises(LambdaTimeoutError):
        runner_process.timeout_handler__(None, None)


def test_lambda_invoker(mocker, runner_process):
    mock_handler = mocker.Mock()
    mock_import_attribute = mocker.patch(
        "smyth.runner.process.import_attribute",
        autospec=True,
        return_value=mock_handler,
    )
    mock_signal = mocker.patch("signal.signal", autospec=True)
    mock_signal.side_effect = lambda signum, frame: None

    mocker.patch.object(runner_process, "output_queue", autospec=True)

    mocker.patch.object(
        runner_process,
        "get_message__",
        autospec=True,
        return_value=[
            {"type": "smyth.lambda.invoke", "event": {"test": "1"}, "context": {}},
            {"type": "smyth.lambda.invoke", "event": {"test": "2"}, "context": {}},
        ],
    )
    mocker.patch.object(
        runner_process,
        "get_event__",
        autospec=True,
        side_effect=[
            {"test": "1"},
            {"test": "2"},
        ],
    )
    mock_get_context__ = mocker.patch.object(
        runner_process, "get_context__", autospec=True
    )
    mock_set_status__ = mocker.patch.object(
        runner_process, "set_status__", autospec=True
    )

    runner_process.lambda_invoker__()
    assert mock_import_attribute.call_count == 1

    mock_set_status__.assert_has_calls(
        [
            mocker.call(SmythHandlerState.COLD),
            mocker.call(SmythHandlerState.WARM),
            mocker.call(SmythHandlerState.WORKING),
            mocker.call(SmythHandlerState.WORKING),
        ]
    )

    assert mock_handler.call_count == 2
    mock_handler.assert_has_calls(
        [
            mocker.call({"test": "1"}, mock_get_context__.return_value),
            mocker.call({"test": "2"}, mock_get_context__.return_value),
        ]
    )



from time import strftime

import pytest

from smyth.runner.fake_context import FakeLambdaContext


def test_fake_lambda_context():
    context = FakeLambdaContext()
    assert context.function_name == "Fake"
    assert context.function_version == "LATEST"
    assert context.invoked_function_arn == "arn:aws:lambda:serverless:Fake"
    assert context.memory_limit_in_mb == "1024"
    assert context.aws_request_id == "1234567890"
    assert context.log_group_name == "/aws/lambda/Fake"
    assert context.log_stream_name == (
        f"{strftime('%Y/%m/%d')}/[$LATEST]58419525dade4d17a495dceeeed44708"
    )


@pytest.mark.parametrize(
    (
        "name",
        "version",
        "timeout",
        "expected_name",
        "expected_version",
        "expected_timeout",
    ),
    [
        ("test", "test", 60, "test", "test", 60),
        ("test", "test", None, "test", "test", 6),
        ("test", None, 120, "test", "LATEST", 120),
        (None, "test", 6, "Fake", "test", 6),
        (None, None, 6, "Fake", "LATEST", 6),
    ],
)
def test_fake_lambda_context_with_params(
    name, version, timeout, expected_name, expected_version, expected_timeout
):
    context = FakeLambdaContext(name=name, version=version, timeout=timeout)
    assert context.function_name == expected_name
    assert context.function_version == expected_version
    assert context.timeout == expected_timeout
    assert context.invoked_function_arn == f"arn:aws:lambda:serverless:{expected_name}"
    assert context.memory_limit_in_mb == "1024"
    assert context.aws_request_id == "1234567890"
    assert context.log_group_name == f"/aws/lambda/{expected_name}"
    assert context.log_stream_name == (
        f"{strftime('%Y/%m/%d')}/[${expected_version}]58419525dade4d17a495dceeeed44708"
    )



import re
from unittest.mock import AsyncMock, Mock

import pytest

from smyth.config import Config, HandlerConfig
from smyth.types import RunnerProcessProtocol, SmythHandler, SmythHandlerState


@pytest.fixture(autouse=True)
def anyio_backend():
    return "asyncio", {"use_uvloop": True}


@pytest.fixture
def smyth_handler(
    mock_lambda_handler,
    mock_event_data_function,
    mock_context_data_function,
    mock_strategy_generator,
):
    return SmythHandler(
        name="test_handler",
        url_path=re.compile(r"/test_handler"),
        lambda_handler_path="tests.conftest.example_handler",
        event_data_function=mock_event_data_function,
        context_data_function=mock_context_data_function,
        strategy_generator=mock_strategy_generator,
    )


@pytest.fixture
def config():
    return Config(
        host="0.0.0.0",
        port=8080,
        handlers={
            "order_handler": HandlerConfig(
                handler_path="tests.conftest.example_handler",
                url_path=r"/test_handler",
            ),
            "product_handler": HandlerConfig(
                handler_path="tests.conftest.example_handler",
                url_path=r"/products/{path:path}",
            ),
        },
        log_level="INFO",
    )


def example_handler(event, context):
    return {"statusCode": 200, "body": "Hello, World!"}


@pytest.fixture
def mock_lambda_handler():
    return example_handler


@pytest.fixture
def mock_event_data_function():
    return AsyncMock()


@pytest.fixture
def mock_context_data_function():
    return AsyncMock()


@pytest.fixture
def mock_strategy_generator():
    return Mock()


@pytest.fixture
def mock_runner_process():
    mock = Mock(
        spec=RunnerProcessProtocol,
    )
    mock.name = "test_process"
    mock.task_counter = 0
    mock.last_used_timestamp = 0
    mock.state = SmythHandlerState.COLD
    return mock



import logging
import logging.config
import os
from enum import Enum
from typing import Annotated, Optional

import typer
import uvicorn
from setproctitle import setproctitle

from smyth.config import get_config, get_config_dict, serialize_config
from smyth.utils import get_logging_config

app = typer.Typer()
config = get_config(get_config_dict())


LOGGER = logging.getLogger(__name__)


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


@app.command()
def run(
    # typer does not handle union types
    smyth_starlette_app: Annotated[
        str,
        typer.Argument(
            help="Path to your SmythStarlette app (or factory with --factory)"
        ),  # noqa: UP007
    ] = "smyth.server.app:create_app",
    factory: Annotated[bool, typer.Option(help="Use factory for app creation")] = True,
    host: Annotated[Optional[str], typer.Option()] = config.host,  # noqa: UP007
    port: Annotated[Optional[int], typer.Option()] = config.port,  # noqa: UP007
    log_level: Annotated[
        Optional[LogLevel],  # noqa: UP007
        typer.Option(
            help=(
                "Override the log level specified in the configuration, "
                "only for the main process"
            )
        ),
    ] = LogLevel(config.log_level),
    quiet: Annotated[
        bool, typer.Option(help="Effectively the same as --log-level=ERROR")
    ] = False,
):
    if host:
        config.host = host
    if port:
        config.port = port
    if log_level:
        config.log_level = log_level.value
    if quiet:
        config.log_level = "ERROR"

    logging_config = get_logging_config(
        log_level=config.log_level, filter_path_prefix=config.smyth_path_prefix
    )

    setproctitle("smyth")
    os.environ["__SMYTH_CONFIG"] = serialize_config(config)

    uvicorn.run(
        smyth_starlette_app,
        factory=factory,
        host=config.host,
        port=config.port,
        reload=True,
        log_config=logging_config,
        timeout_keep_alive=60 * 15,
        lifespan="on",
    )


if __name__ == "__main__":
    app()



import json
import os
from dataclasses import asdict
from pathlib import Path

import pytest

from smyth.config import (
    Config,
    HandlerConfig,
    get_config,
    get_config_dict,
    get_config_file_path,
    serialize_config,
)
from smyth.exceptions import ConfigFileNotFoundError


def test_get_config_file_path():
    assert get_config_file_path() == get_config_file_path("pyproject.toml")

    with pytest.raises(ConfigFileNotFoundError):
        get_config_file_path("not_existing.toml")


def test_get_config_dict(mocker):
    mock_toml = mocker.patch("smyth.config.toml.load")
    mock_get_config_file_path = mocker.patch(
        "smyth.config.get_config_file_path", return_value=Path("pyproject.toml")
    )

    assert get_config_dict() == mock_toml.return_value
    mock_toml.assert_called_once_with(mock_get_config_file_path.return_value)
    mock_get_config_file_path.assert_called_once_with()

    mock_get_config_file_path.reset_mock()
    mock_toml.reset_mock()
    mock_get_config_file_path.return_value = Path("other.toml")

    assert get_config_dict("other.toml") == mock_toml.return_value
    mock_toml.assert_called_once_with(mock_get_config_file_path.return_value)
    mock_get_config_file_path.assert_called_once_with("other.toml")


def test_get_config():
    config_dict = {
        "tool": {
            "smyth": {
                "host": "0.0.0.0",
                "port": 8080,
                "handlers": {
                    "order_handler": {
                        "handler_path": "tests.conftest.example_handler",
                        "url_path": "/test_handler",
                    },
                    "product_handler": {
                        "handler_path": "tests.conftest.example_handler",
                        "url_path": "/products/{path:path}",
                    },
                },
                "log_level": "INFO",
            }
        }
    }

    config = get_config(config_dict)

    assert config.host == "0.0.0.0"
    assert config.port == 8080
    assert config.handlers == {
        "order_handler": HandlerConfig(
            handler_path="tests.conftest.example_handler",
            url_path=r"/test_handler",
        ),
        "product_handler": HandlerConfig(
            handler_path="tests.conftest.example_handler",
            url_path=r"/products/{path:path}",
        ),
    }
    assert config.log_level == "INFO"

    os.environ["__SMYTH_CONFIG"] = serialize_config(config)

    assert get_config(None) == config


def test_serialize_config():
    config = Config(
        host="0.0.0.0",
        port=8080,
        handlers={
            "order_handler": HandlerConfig(
                handler_path="tests.conftest.example_handler",
                url_path=r"/test_handler",
            ),
            "product_handler": HandlerConfig(
                handler_path="tests.conftest.example_handler",
                url_path=r"/products/{path:path}",
            ),
        },
        log_level="INFO",
    )

    assert serialize_config(config) == json.dumps(asdict(config))



from collections.abc import Iterator

from smyth.exceptions import NoAvailableProcessError
from smyth.types import RunnerProcessProtocol, SmythHandlerState


def round_robin(
    handler_name: str, processes: dict[str, list[RunnerProcessProtocol]]
) -> Iterator[RunnerProcessProtocol]:
    """This strategy, not typical for AWS Lambda's behavior, is beneficial
    during development. It rotates among Lambda Processes for each request,
    given that concurrency is set higher than `1`. This approach encourages
    developers to avoid relying on global state across requests, promoting
    best practices in serverless application design.
    """
    while True:
        yield from processes[handler_name]


def first_warm(
    handler_name: str, processes: dict[str, list[RunnerProcessProtocol]]
) -> Iterator[RunnerProcessProtocol]:
    """This strategy prioritizes the use of the first available Lambda Process
    in a "warm" state to handle incoming requests. If no warm instances are
    available, it initiates a "cold start". This behavior more closely mimics
    the operational dynamics of AWS Lambda, where reusing warm instances can
    lead to faster response times."""

    while True:
        warm = None
        cold = None
        for process in processes[handler_name]:
            if process.state == SmythHandlerState.WARM:
                warm = process
            elif process.state == SmythHandlerState.COLD:
                cold = process

        if warm is not None:
            yield warm
        elif cold is not None:
            yield cold
        else:
            raise NoAvailableProcessError("No process available")



import logging
from collections.abc import Callable, Iterable
from datetime import datetime
from importlib import import_module
from logging import LogRecord
from pathlib import Path

from rich.console import Console, ConsoleRenderable, RenderableType
from rich.logging import RichHandler
from rich.table import Table
from rich.text import Text, TextType
from rich.traceback import Traceback

FormatTimeCallable = Callable[[datetime], Text]


def get_logging_config(log_level: str, filter_path_prefix: str | None = None) -> dict:
    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "filters": {},
        "handlers": {
            "console": {
                "class": "smyth.utils.SmythRichHandler",
                "markup": True,
                "rich_tracebacks": True,
                "filters": [],
                "show_path": False,
            },
        },
        "loggers": {
            "smyth": {
                "handlers": ["console"],
                "level": log_level,
                "propagate": False,
            },
            "uvicorn": {
                "handlers": ["console"],
                "level": log_level,
                "propagate": False,
            },
        },
    }
    if filter_path_prefix:
        logging_config["filters"]["smyth_api_filter"] = {  # type: ignore[index]
            "()": "smyth.utils.SmythStatusRouteFilter",
            "smyth_path_prefix": filter_path_prefix,
        }
        logging_config["handlers"]["console"]["filters"].append("smyth_api_filter")  # type: ignore[index]
    return logging_config


class SmythStatusRouteFilter(logging.Filter):
    def __init__(self, name: str = "", smyth_path_prefix: str = "") -> None:
        super().__init__(name)
        self.smyth_path_prefix = smyth_path_prefix

    def filter(self, record):
        return record.getMessage().find(self.smyth_path_prefix) == -1


class LogRender:  # pragma: no cover
    """
    Derived from `rich._log_render.LogRender`.
    """

    def __init__(
        self,
        show_time: bool = True,
        show_level: bool = False,
        show_path: bool = True,
        time_format: str | FormatTimeCallable = "[%X]",
        omit_repeated_times: bool = True,
        level_width: int | None = 8,
    ) -> None:
        self.show_time = show_time
        self.show_level = show_level
        self.show_path = show_path
        self.time_format = time_format
        self.omit_repeated_times = omit_repeated_times
        self.level_width = level_width
        self._last_time: Text | None = None

    def create_header_row(self, record: LogRecord) -> RenderableType:
        issuer = record.name.split(".")[0]

        if issuer == "smyth":
            issuer = "[bold yellow]Smyth[/]"
            process_name = record.processName
        elif issuer == "uvicorn":
            issuer = "[bold blue]Uvicorn[/]"
            process_name = f"Worker[{record.process}]"
        else:
            issuer = issuer.capitalize()
            process_name = record.processName

        return Text.from_markup(
            f"{issuer}:" f"[bold]{process_name}[/]",
            style="log.process",
        )

    def create_time_row(self, log_time, console, time_format) -> RenderableType | None:
        log_time = log_time or console.get_datetime()
        time_format = time_format or self.time_format
        if callable(time_format):
            log_time_display = time_format(log_time)
        else:
            log_time_display = Text(log_time.strftime(time_format))
        if log_time_display == self._last_time and self.omit_repeated_times:
            return Text(" " * len(log_time_display))
        else:
            self._last_time = log_time_display
            return log_time_display

    def create_path_row(
        self, path: str, line_no: int | None, link_path: str | None
    ) -> RenderableType:
        path_text = Text()
        path_text.append(path, style=f"link file://{link_path}" if link_path else "")
        if line_no:
            path_text.append(":")
            path_text.append(
                f"{line_no}",
                style=f"link file://{link_path}#{line_no}" if link_path else "",
            )
        return path_text

    def configure_columns(self, record: LogRecord) -> tuple[bool, bool, bool]:
        full_width = getattr(record, "log_setting", None) == "console_full_width"

        if full_width:
            show_time = False
            show_level = False
            show_path = False
        else:
            show_time = self.show_time
            show_level = self.show_level
            show_path = self.show_path

        return show_time, show_level, show_path

    def __call__(
        self,
        record: LogRecord,
        console: "Console",
        renderables: Iterable["ConsoleRenderable"],
        log_time: datetime | None = None,
        time_format: str | FormatTimeCallable | None = None,
        level: TextType = "",
        path: str | None = None,
        line_no: int | None = None,
        link_path: str | None = None,
    ) -> "Table":
        from rich.containers import Renderables
        from rich.table import Table

        show_time, show_level, show_path = self.configure_columns(record)
        output = Table.grid(padding=(0, 1))
        output.expand = True
        output.add_column(justify="left", min_width=22)
        row: list[RenderableType] = []

        row.append(self.create_header_row(record))

        if show_time:
            output.add_column(style="log.time")
            create_time_row = self.create_time_row(log_time, console, time_format)
            if create_time_row:
                row.append(create_time_row)
        if show_level:
            output.add_column(style="log.level", width=self.level_width)
            row.append(level)

        output.add_column(ratio=1, style="log.message", overflow="fold")
        row.append(Renderables(renderables))

        if show_path and path:
            output.add_column(style="log.path")
            row.append(self.create_path_row(path, line_no, link_path))

        output.add_row(*row)
        return output


class SmythRichHandler(RichHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rich_render = LogRender(
            show_time=True,
            show_level=True,
            show_path=False,
            time_format="[%X]",
            omit_repeated_times=True,
            level_width=8,
        )

    def render(
        self,
        *,
        record: LogRecord,
        traceback: Traceback | None,
        message_renderable: "ConsoleRenderable",
    ) -> "ConsoleRenderable":
        path = Path(record.pathname).name
        level = self.get_level_text(record)
        time_format = None if self.formatter is None else self.formatter.datefmt
        log_time = datetime.fromtimestamp(record.created)

        log_renderable = self.rich_render(
            record=record,
            console=self.console,
            renderables=[message_renderable]
            if not traceback
            else [message_renderable, traceback],
            log_time=log_time,
            time_format=time_format,
            level=level,
            path=path,
            line_no=record.lineno,
            link_path=record.pathname if self.enable_link_path else None,
        )
        return log_renderable


def import_attribute(python_path: str):
    module_name, handler_name = python_path.rsplit(".", 1)
    module = import_module(module_name)
    return getattr(module, handler_name)



from smyth.utils import get_logging_config, import_attribute


def test_get_logging_config():
    logging_config = get_logging_config("DEBUG")
    assert logging_config["version"] == 1
    assert logging_config["disable_existing_loggers"] is False
    assert logging_config["filters"] == {}
    assert (
        logging_config["handlers"]["console"]["class"] == "smyth.utils.SmythRichHandler"
    )
    assert logging_config["handlers"]["console"]["markup"] is True
    assert logging_config["handlers"]["console"]["rich_tracebacks"] is True
    assert logging_config["handlers"]["console"]["filters"] == []
    assert logging_config["handlers"]["console"]["show_path"] is False
    assert logging_config["loggers"]["smyth"]["handlers"] == ["console"]
    assert logging_config["loggers"]["smyth"]["level"] == "DEBUG"
    assert logging_config["loggers"]["smyth"]["propagate"] is False
    assert logging_config["loggers"]["uvicorn"]["handlers"] == ["console"]
    assert logging_config["loggers"]["uvicorn"]["level"] == "DEBUG"
    assert logging_config["loggers"]["uvicorn"]["propagate"] is False

    logging_config = get_logging_config("INFO", "/smyth")
    assert (
        logging_config["filters"]["smyth_api_filter"]["smyth_path_prefix"] == "/smyth"
    )
    assert logging_config["handlers"]["console"]["filters"] == ["smyth_api_filter"]


def test_import_attribute():
    assert import_attribute("smyth.utils.get_logging_config") == get_logging_config



import pytest

from smyth.exceptions import NoAvailableProcessError
from smyth.runner.strategy import first_warm, round_robin
from smyth.types import SmythHandlerState


def test_round_robin():
    strat = round_robin("test_handler", {"test_handler": [1, 2, 3]})
    assert next(strat) == 1
    assert next(strat) == 2
    assert next(strat) == 3
    assert next(strat) == 1
    assert next(strat) == 2


def test_first_warm(mocker):
    mock_cold_process = mocker.Mock()
    mock_cold_process.state = SmythHandlerState.COLD
    mock_working_process = mocker.Mock()
    mock_working_process.state = SmythHandlerState.WORKING
    mock_warm_process = mocker.Mock()
    mock_warm_process.state = SmythHandlerState.WARM

    strat = first_warm(
        "test_handler",
        {"test_handler": [mock_working_process, mock_cold_process, mock_warm_process]},
    )

    assert next(strat) == mock_warm_process
    assert next(strat) == mock_warm_process
    assert next(strat) == mock_warm_process


def test_first_warm_no_warm(mocker):
    mock_cold_process = mocker.Mock()
    mock_cold_process.state = SmythHandlerState.COLD
    mock_working_process = mocker.Mock()
    mock_working_process.state = SmythHandlerState.WORKING
    mock_warm_process = mocker.Mock()
    mock_warm_process.state = SmythHandlerState.WARM

    strat = first_warm(
        "test_handler",
        {"test_handler": [mock_working_process, mock_cold_process, mock_cold_process]},
    )

    assert next(strat) == mock_cold_process
    assert next(strat) == mock_cold_process
    mock_working_process.state = SmythHandlerState.WARM
    assert next(strat) == mock_working_process


def test_first_warm_no_available(mocker):
    mock_working_process = mocker.Mock()
    mock_working_process.state = SmythHandlerState.WORKING

    strat = first_warm(
        "test_handler",
        {"test_handler": [mock_working_process]},
    )

    with pytest.raises(NoAvailableProcessError) as excinfo:
        next(strat)

    assert excinfo.errisinstance(NoAvailableProcessError)



class SmythRuntimeError(Exception):
    """Generic Lambda Runtime exception."""


class ConfigFileNotFoundError(SmythRuntimeError):
    """Config file not found."""


class DispatcherError(SmythRuntimeError):
    pass


class ProcessDefinitionNotFoundError(DispatcherError):
    pass


class NoAvailableProcessError(DispatcherError):
    pass


class DestroyedOnLoadError(DispatcherError):
    pass


class SubprocessError(SmythRuntimeError):
    """Generic subprocess exception."""


class LambdaHandlerLoadError(SubprocessError):
    """Error loading a Lambda handler."""


class LambdaTimeoutError(SubprocessError):
    """Lambda timeout."""


class LambdaInvocationError(SubprocessError):
    """Error invoking a Lambda."""



import re
from unittest.mock import ANY

from smyth.context import generate_context_data
from smyth.types import SmythHandlerState


async def test_generate_context_data(
    smyth_handler,
    mock_runner_process,
):
    assert await generate_context_data(None, smyth_handler, mock_runner_process) == {
        "smyth": {
            "handler": {
                "smyth_handler_config": {
                    "concurrency": 1,
                    "context_data_function": ANY,
                    "event_data_function": ANY,
                    "lambda_handler_path": "tests.conftest.example_handler",
                    "log_level": "INFO",
                    "name": "test_handler",
                    "strategy_generator": ANY,
                    "timeout": None,
                    "url_path": re.compile("/test_handler"),
                },
                "name": "test_handler",
            },
            "process": {
                "last_used_timestamp": 0,
                "name": "test_process",
                "state": SmythHandlerState.COLD,
                "task_counter": 0,
            },
        },
    }



from collections.abc import Awaitable, Callable, Iterator, MutableMapping
from dataclasses import dataclass
from enum import Enum
from re import Pattern
from typing import Any, Protocol

from aws_lambda_powertools.utilities.typing import LambdaContext
from starlette.requests import Request

LambdaEvent = MutableMapping[str, Any]
LambdaHandler = Callable[[LambdaEvent, LambdaContext], dict[str, Any]]
RunnerMessage = MutableMapping[str, Any]
EventDataCallable = Callable[[Request], Awaitable[dict[str, Any]]]
ContextDataCallable = Callable[
    [Request | None, "SmythHandler", "RunnerProcessProtocol"], Awaitable[dict[str, Any]]
]
StrategyGenerator = Callable[
    [str, dict[str, list["RunnerProcessProtocol"]]],
    Iterator["RunnerProcessProtocol"],
]


class SmythHandlerState(str, Enum):
    COLD = "cold"
    WORKING = "working"
    WARM = "warm"


class RunnerProcessProtocol(Protocol):
    name: str
    task_counter: int
    last_used_timestamp: float
    state: SmythHandlerState

    async def asend(self, data) -> RunnerMessage | None: ...

    def stop(self): ...

    def send(self, data) -> RunnerMessage | None: ...

    def is_alive(self) -> bool: ...

    def terminate(self): ...

    def join(self): ...


@dataclass
class SmythHandler:
    name: str
    url_path: Pattern[str]
    lambda_handler_path: str
    event_data_function: EventDataCallable
    context_data_function: ContextDataCallable
    strategy_generator: StrategyGenerator
    timeout: float | None = None
    log_level: str = "INFO"
    concurrency: int = 1



import logging
import logging.config
from collections.abc import Iterator

from starlette.requests import Request
from starlette.routing import compile_path

from smyth.context import generate_context_data
from smyth.event import generate_api_gw_v2_event_data
from smyth.exceptions import ProcessDefinitionNotFoundError
from smyth.runner.process import RunnerProcess
from smyth.runner.strategy import first_warm
from smyth.types import (
    ContextDataCallable,
    EventDataCallable,
    RunnerProcessProtocol,
    SmythHandler,
    StrategyGenerator,
)

LOGGER = logging.getLogger(__name__)


class Smyth:
    smyth_handlers: dict[str, SmythHandler]
    processes: dict[str, list[RunnerProcessProtocol]]
    strategy_generators: dict[str, Iterator["RunnerProcessProtocol"]]

    def __init__(self) -> None:
        self.smyth_handlers = {}
        self.processes = {}
        self.strategy_generators = {}

    def add_handler(
        self,
        name: str,
        path: str,
        lambda_handler_path: str,
        timeout: float | None = None,
        event_data_function: EventDataCallable = generate_api_gw_v2_event_data,
        context_data_function: ContextDataCallable = generate_context_data,
        log_level: str = "INFO",
        concurrency: int = 1,
        strategy_generator: StrategyGenerator = first_warm,
    ):
        self.smyth_handlers[name] = SmythHandler(
            name=name,
            url_path=compile_path(path)[0],
            lambda_handler_path=lambda_handler_path,
            event_data_function=event_data_function,
            context_data_function=context_data_function,
            timeout=timeout,
            log_level=log_level,
            concurrency=concurrency,
            strategy_generator=strategy_generator,
        )

    def __enter__(self):
        self.start_runners()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop_runners()

    def start_runners(self):
        for handler_name, handler_config in self.smyth_handlers.items():
            self.processes[handler_name] = []
            for index in range(handler_config.concurrency):
                process = RunnerProcess(
                    name=f"{handler_name}:{index}",
                    lambda_handler_path=handler_config.lambda_handler_path,
                    log_level=handler_config.log_level,
                )
                process.start()
                LOGGER.info("Started process %s", process.name)
                self.processes[handler_name].append(process)
            self.strategy_generators[handler_name] = handler_config.strategy_generator(
                handler_name, self.processes
            )

    def stop_runners(self):
        for process_group in self.processes.values():
            for process in process_group:
                LOGGER.info("Stopping process %s", process.name)
                process.stop()
        for process_group in self.processes.values():
            for process in process_group:
                LOGGER.debug("Joining process %s", process.name)
                if process.is_alive():
                    process.terminate()
                    process.join()

    def get_handler_for_request(self, path: str) -> SmythHandler:
        for handler_def in self.smyth_handlers.values():
            if handler_def.url_path.match(path):
                return handler_def
        raise ProcessDefinitionNotFoundError(
            f"No process definition found for path {path}"
        )

    def get_handler_for_name(self, name: str) -> SmythHandler:
        return self.smyth_handlers[name]

    async def dispatch(
        self,
        smyth_handler: SmythHandler,
        request: Request,
        event_data_function: EventDataCallable | None = None,
    ):
        """
        Smyth.dispatch is used upon a request that would normally be formed by an
        AWS trigger. It is responsible for finding the appropriate process
        for the request, invoking the process, and translating the response
        """
        process = next(self.strategy_generators[smyth_handler.name])
        if process is None:
            raise ProcessDefinitionNotFoundError(
                f"No process definition found for handler {smyth_handler.name}"
            )

        if event_data_function is None:
            event_data_function = smyth_handler.event_data_function

        event_data = await event_data_function(request)
        context_data = await smyth_handler.context_data_function(
            request, smyth_handler, process
        )

        return await process.asend(
            {
                "type": "smyth.lambda.invoke",
                "event": event_data,
                "context": context_data,
            }
        )

    async def invoke(self, handler: SmythHandler, event_data: dict):
        """
        Smyth.invoke is used to invoke a handler directly, without going through
        Starlette or when a direct invocation is needed (e.g., when invoking
        a lambda with boto3) - on direct invocation the event holds only the data
        passed in the invokation. There's no Starlette request involved.
        """
        process = next(self.strategy_generators[handler.name])
        if process is None:
            raise ProcessDefinitionNotFoundError(
                f"No process definition found for handler {handler.name}"
            )
        context_data = await handler.context_data_function(None, handler, process)
        return await process.asend(
            {
                "type": "smyth.lambda.invoke",
                "event": event_data,
                "context": context_data,
            }
        )



import json
import os
from dataclasses import asdict, dataclass, field
from pathlib import Path

import toml

from smyth.exceptions import ConfigFileNotFoundError


@dataclass
class HandlerConfig:
    handler_path: str
    url_path: str
    timeout: float | None = None
    event_data_function_path: str = "smyth.event.generate_api_gw_v2_event_data"
    context_data_function_path: str = "smyth.context.generate_context_data"
    log_level: str = "DEBUG"
    concurrency: int = 1
    strategy_generator_path: str = "smyth.runner.strategy.first_warm"


@dataclass
class Config:
    host: str = "0.0.0.0"
    port: int = 8080
    handlers: dict[str, HandlerConfig] = field(default_factory=dict)
    log_level: str = "INFO"
    smyth_path_prefix: str = "/smyth"

    @classmethod
    def from_dict(cls, config_dict: dict):
        handler_data = config_dict.pop("handlers")
        handlers = {
            handler_name: HandlerConfig(**handler_config)
            for handler_name, handler_config in handler_data.items()
        }
        return cls(**config_dict, handlers=handlers)


def get_config_file_path(file_name: str = "pyproject.toml") -> Path:
    """Get config file path. If not found raise exception."""
    directory = Path.cwd()
    while not directory.joinpath(file_name).exists():
        if directory == directory.parent:
            raise ConfigFileNotFoundError(f"Config file {file_name} not found.")
        directory = directory.parent
    return directory.joinpath(file_name).resolve()


def get_config_dict(config_file_name: str | None = None) -> dict:
    """Get config dict."""
    if config_file_name:
        config_file_path = get_config_file_path(config_file_name)
    else:
        config_file_path = get_config_file_path()

    return toml.load(config_file_path)


def get_config(config_dict: dict) -> Config:
    """Get config."""
    if environ_config := os.environ.get("__SMYTH_CONFIG"):
        config_data = json.loads(environ_config)
        return Config.from_dict(config_data)
    return Config.from_dict(config_dict["tool"]["smyth"])


def serialize_config(config: Config) -> str:
    return json.dumps(asdict(config))



from dataclasses import asdict
from typing import Any

from starlette.requests import Request

from smyth.types import RunnerProcessProtocol, SmythHandler


async def generate_context_data(
    request: Request | None, smyth_handler: SmythHandler, process: RunnerProcessProtocol
):
    """
    The data returned by this function is passed to the
    `smyth.runner.FaneContext` as kwargs.
    """
    context: dict[str, Any] = {
        "smyth": {
            "process": {
                "name": process.name,
                "state": process.state,
                "task_counter": process.task_counter,
                "last_used_timestamp": process.last_used_timestamp,
            },
            "handler": {
                "name": smyth_handler.name,
                "smyth_handler_config": asdict(smyth_handler),
            },
        }
    }
    if smyth_handler.timeout is not None:
        context["timeout"] = smyth_handler.timeout
    return context



import pytest

from smyth.event import (
    generate_api_gw_v2_event_data,
    generate_lambda_invocation_event_data,
)

pytestmark = pytest.mark.anyio


async def test_generate_api_gw_v2_event_data(mocker):
    mock_request = mocker.Mock()
    mock_request.body = mocker.AsyncMock(return_value=b"")
    mock_request.headers = {}
    mock_request.query_params = {}
    mock_request.client.host = "127.0.0.1"
    mock_request.method = "GET"
    mock_request.url.path = "/test"
    mock_request.url.query = ""
    mock_request.url.scheme = "http"

    assert await generate_api_gw_v2_event_data(mock_request) == {
        "version": "2.0",
        "rawPath": "/test",
        "body": "",
        "isBase64Encoded": False,
        "headers": {},
        "queryStringParameters": {},
        "requestContext": {
            "http": {
                "method": "GET",
                "path": "/test",
                "protocol": "http",
                "sourceIp": "127.0.0.1",
            },
            "routeKey": "GET /test",
            "accountId": "offlineContext_accountId",
            "stage": "$default",
        },
        "routeKey": "GET /test",
        "rawQueryString": "",
    }


async def test_generate_lambda_invokation_event_data(mocker):
    mock_request = mocker.Mock()
    mock_request.json = mocker.AsyncMock(return_value={"test": "test"})

    assert await generate_lambda_invocation_event_data(mock_request) == {"test": "test"}



import sys
from time import strftime, time

from aws_lambda_powertools.utilities.typing import LambdaContext


class FakeLambdaContext(LambdaContext):
    def __init__(
        self,
        name: str | None = None,
        version: str | None = "LATEST",
        timeout: int | None = None,
        **kwargs,
    ):
        if name is None:
            name = "Fake"
        self.name = name

        if version is None:
            version = "LATEST"
        self.version = version

        self.created = time()

        if timeout is None:
            timeout = 6
        self.timeout = timeout

        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_remaining_time_in_millis(self) -> int:  # type: ignore[override]
        return int(
            max(
                (self.timeout * 1000)
                - (int(round(time() * 1000)) - int(round(self.created * 1000))),
                0,
            )
        )

    @property
    def function_name(self):
        return self.name

    @property
    def function_version(self):
        return self.version

    @property
    def invoked_function_arn(self):
        return "arn:aws:lambda:serverless:" + self.name

    @property
    def memory_limit_in_mb(self):
        return "1024"

    @property
    def aws_request_id(self):
        return "1234567890"

    @property
    def log_group_name(self):
        return "/aws/lambda/" + self.name

    @property
    def log_stream_name(self):
        return (
            strftime("%Y/%m/%d")
            + "/[$"
            + self.version
            + "]58419525dade4d17a495dceeeed44708"
        )

    @property
    def log(self):
        return sys.stdout.write



from starlette.requests import Request


async def generate_api_gw_v2_event_data(request: Request):
    source_ip = None
    if request.client:
        source_ip = request.client.host
    return {
        "version": "2.0",
        "rawPath": request.url.path,
        "body": (await request.body()).decode("utf-8"),
        "isBase64Encoded": False,
        "headers": dict(request.headers),
        "queryStringParameters": dict(request.query_params),
        "requestContext": {
            "http": {
                "method": request.method,
                "path": request.url.path,
                "protocol": request.url.scheme,
                "sourceIp": source_ip,
            },
            "routeKey": f"{request.method} {request.url.path}",
            "accountId": "offlineContext_accountId",
            "stage": "$default",
        },
        "routeKey": f"{request.method} {request.url.path}",
        "rawQueryString": request.url.query,
    }


async def generate_lambda_invocation_event_data(request: Request):
    return await request.json()



import inspect
import logging
import logging.config
import signal
import sys
import traceback
from multiprocessing import Process, Queue, set_start_method
from queue import Empty
from time import time

from asgiref.sync import sync_to_async
from setproctitle import setproctitle

from smyth.exceptions import (
    LambdaHandlerLoadError,
    LambdaInvocationError,
    LambdaTimeoutError,
    SubprocessError,
)
from smyth.runner.fake_context import FakeLambdaContext
from smyth.types import LambdaHandler, RunnerMessage, SmythHandlerState
from smyth.utils import get_logging_config, import_attribute

set_start_method("spawn", force=True)
LOGGER = logging.getLogger(__name__)


class RunnerProcess(Process):
    name: str
    task_counter: int
    last_used_timestamp: float
    state: SmythHandlerState

    def __init__(self, name: str, lambda_handler_path: str, log_level: str = "INFO"):
        self.name = name
        self.task_counter = 0
        self.last_used_timestamp = 0
        self.state = SmythHandlerState.COLD

        self.input_queue: Queue[RunnerMessage] = Queue(maxsize=1)
        self.output_queue: Queue[RunnerMessage] = Queue(maxsize=1)

        self.lambda_handler_path = lambda_handler_path
        self.log_level = log_level
        super().__init__(
            name=name,
        )

    def stop(self):
        self.input_queue.put({"type": "smyth.stop"})
        self.join()
        self.input_queue.close()
        self.output_queue.close()
        self.input_queue.join_thread()
        self.output_queue.join_thread()

    def send(self, data) -> RunnerMessage | None:
        LOGGER.debug("Sending data to process %s: %s", self.name, data)
        self.task_counter += 1
        self.last_used_timestamp = time()
        self.input_queue.put(data)

        while True:
            if not self.is_alive():
                LOGGER.error(
                    "Process is not alive, this should generally not happen. "
                    "Restart Smyth and check your configuration. "
                    "This often happens when the handler can't be loaded "
                    "(i.e. an exception is raised when importing the handler)."
                )
                raise SubprocessError("Process is not alive")
            try:
                message = self.output_queue.get(block=True, timeout=1)
            except Empty:
                continue
            except Exception as error:
                LOGGER.error("Error getting message from output queue: %s", error)
                return None

            LOGGER.debug("Received message from process %s: %s", self.name, message)
            if message["type"] == "smyth.lambda.status":
                self.state = SmythHandlerState(message["status"])
            elif message["type"] == "smyth.lambda.response":
                self.state = SmythHandlerState.WARM
                return message["response"]
            elif message["type"] == "smyth.lambda.error":
                self.state = SmythHandlerState.WARM
                if message["response"]["type"] == "LambdaTimeoutError":
                    raise LambdaTimeoutError(message["response"]["message"])
                else:
                    raise LambdaInvocationError(message["response"]["message"])

    @sync_to_async(thread_sensitive=False)
    def asend(self, data) -> RunnerMessage | None:
        return self.send(data)

    # Backend

    def run(self):
        setproctitle(f"smyth:{self.name}")
        logging.config.dictConfig(get_logging_config(self.log_level))
        self.lambda_invoker__()

    def get_message__(self):
        while True:
            try:
                message = self.input_queue.get(block=True, timeout=1)
            except KeyboardInterrupt:
                LOGGER.debug("Stopping process")
                return
            except Empty:
                continue
            else:
                LOGGER.debug("Received message: %s", message)
                if message["type"] == "smyth.stop":
                    LOGGER.debug("Stopping process")
                    return
                yield message

    def get_event__(self, message):
        return message["event"]

    def get_context__(self, message):
        return FakeLambdaContext(**message["context"])

    def import_handler__(self, lambda_handler_path, event, context):
        LOGGER.info("Starting cold, importing '%s'", lambda_handler_path)
        try:
            handler = import_attribute(lambda_handler_path)
        except ImportError as error:
            raise LambdaHandlerLoadError(
                f"Error importing handler: {error}, module not found"
            ) from error
        except AttributeError as error:
            raise LambdaHandlerLoadError(
                f"Error importing handler: {error}, attribute in module not found"
            ) from error

        sig = inspect.signature(handler)
        try:
            sig.bind(event, context)
        except TypeError:
            LOGGER.warning(
                "Handler signature does not match event and context, "
                "using `event` and `context` as parameters."
            )
        return handler

    def set_status__(self, status: SmythHandlerState):
        self.output_queue.put({"type": "smyth.lambda.status", "status": status})

    @staticmethod
    def timeout_handler__(signum, frame):
        raise LambdaTimeoutError("Lambda timeout")

    def lambda_invoker__(self):
        sys.stdin = open("/dev/stdin")
        lambda_handler: LambdaHandler | None = None
        self.set_status__(SmythHandlerState.COLD)

        for message in self.get_message__():
            if message.get("type") != "smyth.lambda.invoke":
                LOGGER.error("Invalid message type: %s", message.get("type"))
                continue

            event = self.get_event__(message)
            context = self.get_context__(message)

            if not lambda_handler:
                lambda_handler = self.import_handler__(
                    self.lambda_handler_path,
                    event,
                    context,
                )
                self.set_status__(SmythHandlerState.WARM)

            signal.signal(signal.SIGALRM, self.timeout_handler__)
            signal.alarm(int(context.timeout))
            self.set_status__(SmythHandlerState.WORKING)
            try:
                response = lambda_handler(event, context)
            except Exception as error:
                LOGGER.exception(
                    "Error invoking lambda: %s",
                    error,
                    extra={"log_setting": "console_full_width"},
                )
                self.output_queue.put(
                    {
                        "type": "smyth.lambda.error",
                        "response": {
                            "type": type(error).__name__,
                            "message": str(error),
                            "stacktrace": traceback.format_exc(),
                        },
                    }
                )
            else:
                self.output_queue.put(
                    {
                        "type": "smyth.lambda.response",
                        "response": response,
                    }
                )
            finally:
                signal.alarm(0)



import logging
from typing import Any

from starlette import status
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from smyth.event import generate_lambda_invocation_event_data
from smyth.exceptions import LambdaInvocationError, LambdaTimeoutError, SubprocessError
from smyth.smyth import Smyth
from smyth.types import EventDataCallable, SmythHandler

LOGGER = logging.getLogger(__name__)


async def dispatch(
    smyth: Smyth,
    smyth_handler: SmythHandler,
    request: Request,
    event_data_function: EventDataCallable | None = None,
):
    """
    Dispatches a request to Smyth and translates a Smyth
    response to a Starlette response.
    """
    try:
        result = await smyth.dispatch(
            smyth_handler, request, event_data_function=event_data_function
        )
    except LambdaInvocationError as error:
        return Response(str(error), status_code=status.HTTP_502_BAD_GATEWAY)
    except LambdaTimeoutError:
        return Response("Lambda timeout", status_code=status.HTTP_408_REQUEST_TIMEOUT)
    except SubprocessError as error:
        return Response(str(error), status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

    if not result:
        return Response(
            "No response", status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
        )

    return Response(
        content=result.get("body", result),
        status_code=result.get("statusCode", 200),
        headers=result.get("headers", {}),
    )


async def lambda_invoker_endpoint(request: Request):
    smyth: Smyth = request.app.smyth
    smyth_handler = smyth.get_handler_for_request(request.url.path)
    return await dispatch(smyth, smyth_handler, request)


async def invocation_endpoint(request: Request):
    smyth: Smyth = request.app.smyth
    function = request.path_params["function"]
    try:
        smyth_handler = smyth.get_handler_for_name(function)
    except KeyError:
        return Response(
            f"Function {function} not found", status_code=status.HTTP_404_NOT_FOUND
        )
    smyth_handler.event_data_function = generate_lambda_invocation_event_data
    return await dispatch(
        smyth,
        smyth_handler,
        request,
        event_data_function=generate_lambda_invocation_event_data,
    )


async def status_endpoint(request: Request):
    smyth: Smyth = request.app.smyth

    response_data: dict[str, Any] = {
        "lambda handlers": {},
    }

    for process_group_name, process_group in smyth.processes.items():
        response_data["lambda handlers"][process_group_name] = {  # type: ignore[index]
            "processes": [],
        }
        for process in process_group:
            response_data["lambda handlers"][process_group_name]["processes"].append(  # type: ignore[index]
                {
                    "state": process.state,
                    "task_counter": process.task_counter,
                }
            )
    return JSONResponse(
        content=response_data,
        status_code=status.HTTP_200_OK,
    )
